// Copyright 2017 Fortanix, Inc. All rights reserved.
//
// This example demonstrates the performance of encryption and decryption
// using the Fortanix SDKMS batch crypto APIs.
//
// This test creates 100 AES-256 keys. It then performs batch encyption using
// randomly selected keys on randomly generated data.

package com.fortanix.sdkms.v1.examples;

import java.util.*;

import org.apache.commons.cli.*;

import com.fortanix.sdkms.v1.ApiClient;
import com.fortanix.sdkms.v1.ApiException;
import com.fortanix.sdkms.v1.Configuration;
import com.fortanix.sdkms.v1.api.AuthenticationApi;
import com.fortanix.sdkms.v1.api.EncryptionAndDecryptionApi;
import com.fortanix.sdkms.v1.api.SecurityObjectsApi;
import com.fortanix.sdkms.v1.auth.ApiKeyAuth;
import com.fortanix.sdkms.v1.model.AuthResponse;
import com.fortanix.sdkms.v1.model.BatchDecryptRequest;
import com.fortanix.sdkms.v1.model.BatchDecryptRequestInner;
import com.fortanix.sdkms.v1.model.BatchDecryptResponse;
import com.fortanix.sdkms.v1.model.BatchEncryptRequest;
import com.fortanix.sdkms.v1.model.BatchEncryptRequestInner;
import com.fortanix.sdkms.v1.model.BatchEncryptResponse;
import com.fortanix.sdkms.v1.model.BatchEncryptResponseInner;
import com.fortanix.sdkms.v1.model.CryptMode;
import com.fortanix.sdkms.v1.model.DecryptRequest;
import com.fortanix.sdkms.v1.model.EncryptRequest;
import com.fortanix.sdkms.v1.model.KeyObject;
import com.fortanix.sdkms.v1.model.ObjectType;
import com.fortanix.sdkms.v1.model.SobjectRequest;

// For convenience, you might want to just use the following imports instead:
//import com.fortanix.sdkms.v1.*;
//import com.fortanix.sdkms.v1.api.*;
//import com.fortanix.sdkms.v1.auth.*;
//import com.fortanix.sdkms-v1.models.*;

enum Phase { CREATE_KEYS, ENCRYPT, DECRYPT, CLEAN_UP }

class Parameters {
    public boolean debug;
    public boolean debugTls;
    public String apiKey;
    public String server;
    public String truststore;
    public int numThreads = 1;
    public int numKeys = 100;
    public int batchSize = 100;
    public int requestSize = 48;
    public int numIters = 100;

    public Parameters() {

    }

    public String toString() {
        return "num keys " + numKeys + " batch size " + batchSize +
            " request size " + requestSize + " iterations " + numIters;
    }
}


class BatchThread implements Runnable {
    Parameters parameters;
    private String myId;
    private ArrayList<String> keyIds;
    private ArrayList<BatchEncryptResponse> ciphertexts;
    private Phase phase;

    public BatchThread(Parameters parameters) {
        keyIds = new ArrayList<String>();
        ciphertexts = new ArrayList<BatchEncryptResponse>();
        myId = randomId();

        this.parameters = parameters;
        connect();
    }

    public void setPhase(Phase phase) {
        this.phase = phase;
    }

    private String randomId() {
        Random rand = new Random();
        byte[] bytes = new byte[64];
        rand.nextBytes(bytes);
        return Base64.getEncoder().encodeToString(bytes);
    }

    private String randomKeyId() {
        Random rand = new Random();
        return keyIds.get(rand.nextInt(parameters.numKeys));
    }

    private byte[] randomData() {
        Random rand = new Random();
        byte[] bytes = new byte[parameters.requestSize];
        rand.nextBytes(bytes);
        return bytes;
    }
    
    private void connect() {
        ApiClient client = new ApiClient();

        // Set the name of the server to talk to.
        client.setBasePath(parameters.server);

        // This optionally enables verbose logging in the API library.
        client.setDebugging(parameters.debug);

        // The default ApiClient (and its configured authorization) will be
        // used for constructing the specific API objects, such as
        // EncryptionAndDecryptionApi and SecurityObjectsApi.
        Configuration.setDefaultApiClient(client);

        // If you need a trust store for the server's certificate, you can
        // configure it here or via -Djavax.net.ssl.trustStore= on the
        // java command line.
        if (parameters.truststore != null) {
            System.setProperty("javax.net.ssl.trustStore",
                               parameters.truststore);
        }

        // This optionally enables very verbose logging in the Java network
        // libraries.
        if (parameters.debugTls) {
            System.setProperty("javax.net.debug", "all");
        }

        client.setBasicAuthString(parameters.apiKey);

        // Acquire a bearer token to use for other APIs.
        try {
            AuthResponse response = new AuthenticationApi().authorize();
            String bearerToken = response.getAccessToken();
            if (parameters.debug) {
                System.err.printf("Received Bearer token %s\n", bearerToken);
            }

            // Configure the client library to use the bearer token.
            ApiKeyAuth bearerAuth =
                (ApiKeyAuth) client.getAuthentication("bearerToken");
            bearerAuth.setApiKey(bearerToken);
            bearerAuth.setApiKeyPrefix("Bearer");
        } catch (ApiException e) {
            System.err.println("Unable to authenticate: " + e.getMessage());
            System.exit(1);
        }
    }

    private void createKeys() throws ApiException {
        System.out.println("Creating keys...");
        long start = System.currentTimeMillis();
        SecurityObjectsApi api = new SecurityObjectsApi();
        for (int i = 0; i < parameters.numKeys; i++) {
            SobjectRequest createKeyRequest = new SobjectRequest()
                .objType(ObjectType.AES)
                .keySize(256)
                .name("batch-test-" + myId + Integer.toString(i));
            KeyObject key = api.generateSecurityObject(createKeyRequest);
            
            keyIds.add(key.getKid());
        }
        long end = System.currentTimeMillis();
        System.out.printf("Created %d keys in %d milliseconds\n",
                          parameters.numKeys, end - start);
        System.out.printf("%g keys created per second\n",
                          parameters.numKeys * 1000 / (double) (end - start));
    }

    private void measureBatchEncryptPerformance() throws ApiException {
        System.out.println("Measuring encrypt performance...");
        System.out.println(parameters);
        long start = System.currentTimeMillis();
        for (int iter = 0; iter < parameters.numIters; iter++) {
            BatchEncryptRequest batch = new BatchEncryptRequest();

            for (int i = 0; i < parameters.batchSize; i++) {
                EncryptRequest request = new EncryptRequest()
                    .alg(ObjectType.AES)
                    .mode(CryptMode.CBC)
                    .plain(randomData());
                BatchEncryptRequestInner inner = new BatchEncryptRequestInner()
                    .kid(randomKeyId())
                    .request(request);
                batch.add(inner);
            }

            EncryptionAndDecryptionApi api = new EncryptionAndDecryptionApi();
            BatchEncryptResponse response = api.batchEncrypt(batch);
            ciphertexts.add(response);
        }
        long end = System.currentTimeMillis();
        System.out.printf("Average encryption latency: %gms\n",
                          (end - start) / (double) parameters.numIters /
                          parameters.batchSize);
        System.out.printf("Encryptions per second: %g\n",
                          parameters.numIters * parameters.batchSize * 1000 /
                          (double) (end - start));
    }

    private void measureBatchDecryptPerformance() throws ApiException {
        System.out.println("Measuring decrypt performance...");
        long start = System.currentTimeMillis();
        for (int iter = 0; iter < parameters.numIters; iter++) {
            BatchDecryptRequest batch = new BatchDecryptRequest();

            for (int i = 0; i < parameters.batchSize; i++) {
                BatchEncryptResponse cipher = ciphertexts.get(iter);
                DecryptRequest request = new DecryptRequest()
                    .alg(ObjectType.AES)
                    .mode(CryptMode.CBC)
                    .cipher(cipher.get(i).getBody().getCipher())
                    .iv(cipher.get(i).getBody().getIv());
                BatchDecryptRequestInner inner = new BatchDecryptRequestInner()
                    .kid(cipher.get(i).getBody().getKid())
                    .request(request);
                batch.add(inner);
            }
            EncryptionAndDecryptionApi api = new EncryptionAndDecryptionApi();
            BatchDecryptResponse response = api.batchDecrypt(batch);
        }
        long end = System.currentTimeMillis();
        System.out.printf("Average decryption latency: %gms\n",
                          (end - start) / (double) parameters.numIters /
                          parameters.batchSize);
        System.out.printf("Decryptions per second: %g\n",
                          parameters.numIters * parameters.batchSize * 1000 /
                          (double) (end - start));
    }
    
    private void removeKeys() {
        SecurityObjectsApi api = new SecurityObjectsApi();
        for (String kid : keyIds) {
            try {
                api.deleteSecurityObject(kid);
            } catch (ApiException e) {
                // Ignore failures.
            }
        }
    }

    public void run() {
        try {
            switch (phase) {
            case CREATE_KEYS:
                createKeys();
                break;
            case ENCRYPT:
                measureBatchEncryptPerformance();
                break;
            case DECRYPT:
                measureBatchDecryptPerformance();
                break;
            case CLEAN_UP:
                connect();
                removeKeys();
                break;
            }
        } catch (ApiException e) {
            // Ignore failures.
        }
    }
}

public class BatchCryptoPerf {
    private ArrayList<BatchThread> threadObjects;
    private ArrayList<Thread> threads;
    private Parameters parameters;

    public void run() {
        for (int i = 0; i < parameters.numThreads; i++) {
            BatchThread threadObject = new BatchThread(parameters);
            threadObjects.add(threadObject);
        }

        Phase[] phases = { Phase.CREATE_KEYS, Phase.ENCRYPT, Phase.DECRYPT,
                           Phase.CLEAN_UP };
        for (Phase phase : phases) {
            threads = new ArrayList<Thread>();
            for (int i = 0; i < parameters.numThreads; i++) {
                BatchThread threadObject = threadObjects.get(i);
                threadObject.setPhase(phase);
                threads.add(new Thread(threadObject));
            }
            for (int i = 0; i < parameters.numThreads; i++) {
                threads.get(i).start();
            }
            for (int i = 0; i < parameters.numThreads; i++) {
                try {
                    threads.get(i).join();
                } catch (InterruptedException e) {
                    // ignore
                }
            }
        }
    }

    protected void usage(Options options) {
        HelpFormatter formatter = new HelpFormatter();
        formatter.printHelp("BatchCryptoPerf",
                            "This program demonstrates large batches " +
                            "of cryptographic operations", options, "");
    }
    
    public void parseCmdline(String[] args) {
        Options options = new Options();
        options.addOption(Option.builder("s")
                          .longOpt("server")
                          .hasArg()
                          .type(String.class)
                          .desc("SDKMS server endpoint (e.g. " +
                                "https://apps.smartkey.io")
                          .build());

        options.addOption(Option.builder("k")
                          .longOpt("api-key")
                          .required(true)
                          .hasArg()
                          .type(String.class)
                          .desc("API key for this application")
                          .build());
        options.addOption(Option.builder("t")
                          .longOpt("truststore")
                          .hasArg()
                          .type(String.class)
                          .desc("Java truststore containing server certificate")
                          .build());
        options.addOption(Option.builder("n")
                          .longOpt("num-threads")
                          .hasArg()
                          .type(Number.class)
                          .desc("Number of client threads to run")
                          .build());
        options.addOption(Option.builder("b")
                          .longOpt("batch-size")
                          .hasArg()
                          .type(Number.class)
                          .desc("Number of requests in each batch")
                          .build());
        options.addOption(Option.builder("y")
                          .longOpt("num-keys")
                          .hasArg()
                          .type(Number.class)
                          .desc("Number of keys to create")
                          .build());
        options.addOption(Option.builder("r")
                          .longOpt("request-size")
                          .hasArg()
                          .type(Number.class)
                          .desc("Data size for each encrypt/decrypt operation")
                          .build());
        options.addOption(Option.builder("i")
                          .longOpt("num-iters")
                          .hasArg()
                          .type(Number.class)
                          .desc("Number of iterations to run")
                          .build());
        options.addOption(Option.builder("d")
                          .longOpt("debug")
                          .desc("Enable client API library debugging")
                          .build());
        options.addOption(Option.builder("D")
                          .longOpt("debug-tls")
                          .desc("Enable Java TLS debugging")
                          .build());
        options.addOption(Option.builder("h")
                          .longOpt("help")
                          .desc("Get usage information and exit")
                          .build());
        CommandLineParser parser = new DefaultParser();
        try {
            CommandLine cmdline = parser.parse(options, args);
            if (cmdline.hasOption("help")) {
                usage(options);
                System.exit(0);
            }
            parameters.debug = cmdline.hasOption("debug");
            parameters.debugTls = cmdline.hasOption("debug-tls");
            parameters.apiKey = cmdline.getOptionValue("api-key");

            if (cmdline.hasOption("server")) {
                parameters.server = cmdline.getOptionValue("server");
            } else {
                parameters.server = "https://apps.smartkey.io";
            }
            if (cmdline.hasOption("truststore")) {
                parameters.truststore = cmdline.getOptionValue("truststore");
            }
            if (cmdline.hasOption("num-threads")) {
                parameters.numThreads = ((Number) cmdline.getParsedOptionValue(
                                             "num-threads")).intValue();
            }
            if (cmdline.hasOption("batch-size")) {
                parameters.batchSize = ((Number) cmdline.getParsedOptionValue(
                                            "batch-size")).intValue();
            }
            if (cmdline.hasOption("num-keys")) {
                parameters.numKeys = ((Number) cmdline.getParsedOptionValue(
                                          "num-keys")).intValue();
            }
            if (cmdline.hasOption("request-size")) {
                parameters.requestSize = ((Number) cmdline.getParsedOptionValue(
                                              "request-size")).intValue();
            }
            if (cmdline.hasOption("num-iters")) {
                parameters.numIters = ((Number) cmdline.getParsedOptionValue(
                                           "num-iters")).intValue();
            }
        } catch (ParseException e) {
            usage(options);
            System.exit(1);
        }
    }

    public BatchCryptoPerf() {
        threadObjects = new ArrayList<BatchThread>();
        parameters = new Parameters();
    }

    public static void main(String[] args) {
        BatchCryptoPerf crypto = new BatchCryptoPerf();
        crypto.parseCmdline(args);
        crypto.run();
    }
}
