// Copyright 2017 Fortanix, Inc. All rights reserved.
//
// This example demonstrates how to connect to the SDKMS server as an
// application and authenticate using a client certificate (rather than
// an API key).
//

package com.fortanix.sdkms.v1.examples;

import java.util.List;

import org.apache.commons.cli.*;

// API library imports.
import com.fortanix.sdkms.v1.ApiClient;
import com.fortanix.sdkms.v1.ApiException;
import com.fortanix.sdkms.v1.Configuration;
import com.fortanix.sdkms.v1.api.AuthenticationApi;
import com.fortanix.sdkms.v1.api.SecurityObjectsApi;
import com.fortanix.sdkms.v1.auth.ApiKeyAuth;
import com.fortanix.sdkms.v1.model.AuthResponse;
import com.fortanix.sdkms.v1.model.KeyObject;

// For convenience, you might want to just use the following imports instead:
//import com.fortanix.sdkms.v1.*;
//import com.fortanix.sdkms.v1.api.*;
//import com.fortanix.sdkms.v1.auth.*;
//import com.fortanix.sdkms.v1.models.*;

public class ClientCertificate {
    private boolean debug;
    private boolean debug_tls;
    private String app_id;
    private String server;
    private String keystore;
    private String keystore_password;
    private String truststore;

    ClientCertificate() {}

    public void connect() throws ApiException {

        ApiClient client = new ApiClient();
        // This optionally enables verbose logging in the API library.
        client.setDebugging(debug);

        // Configure the Java system properties to enable the use of your
        // client certificate.
        System.setProperty("javax.net.ssl.keyStoreType", "pkcs12");
        System.setProperty("javax.net.ssl.keyStore", keystore);
        System.setProperty("javax.net.ssl.keyStorePassword", keystore_password);

        // If you need a trust store for the server's certificate, you can
        // configure it here, or via -Djavax.net.ssl.trustStore= on the
        // java command line.
        if (truststore != null) {
            System.setProperty("javax.net.ssl.trustStore", truststore);
        }

        // This optionally enables very verbose logging in the Java network
        // libraries. This can be particularly useful in debugging client
        // certificate authentication.
        if (debug_tls) {
            System.setProperty("javax.net.debug", "all");
        }

        // When authenticating as an application, the username is the
        // application's ID, and the password is the application's
        // secret. Since we are authenticating with a client certificate
        // instead of an application secret, we must set the user name but
        // not the password.
        client.setUsername(app_id);

        // The SDKMS server instance to connect to.
        client.setBasePath(server);

        // The default ApiClient (and its configured authorization) will be
        // used for constructing the specific API objects, such as
        // EncryptionAndDecryptionApi and SecurityObjectsApi.
        Configuration.setDefaultApiClient(client);

        // Call the AuthenticationApi endpoint authorize() using our app's
        // ID and the client certificate to get a bearer token.
        AuthenticationApi authApi = new AuthenticationApi();
        AuthResponse response = authApi.authorize();

        // Configure the library to use the acquired bearer token to
        // authenticate to APIs that need token-based authorization.
        ApiKeyAuth auth = (ApiKeyAuth) client.getAuthentication("bearerToken");
        auth.setApiKey(response.getAccessToken());
        auth.setApiKeyPrefix("Bearer");
    }

    protected void listSecurityObjects() {
        // This call will implicitly use the bearer token to authorize.
        SecurityObjectsApi sobjectsApi = new SecurityObjectsApi();
        // The null parameters here are for filtering returned objects by name,
        // group, creator app ID or creator user ID, respectively. We're just
        // getting all keys here.
        try {
            List<KeyObject> sobjects =
                sobjectsApi.getSecurityObjects(null, null, null);

            System.out.println("Found " + sobjects.size() + " security objects");
            for (KeyObject sobject : sobjects) {
                System.out.println(sobject.getName() + ": " + sobject.getObjType());
            }
        } catch (ApiException e) {
            System.out.println("Unable to list security objects" +
                               e.getMessage());
            System.exit(1);
        }
    }

    protected void usage(Options options) {
        HelpFormatter formatter = new HelpFormatter();
        formatter.printHelp("ClientCertificate", options);
    }

    protected void parseCmdline(String[] args) {
        Options options = new Options();
        options.addOption(Option.builder("a")
                          .longOpt("app-id")
                          .required(true)
                          .hasArg()
                          .type(String.class)
                          .desc("SDKMS application ID to authorize as")
                          .build());

        options.addOption(Option.builder("s")
                          .longOpt("server")
                          .hasArg()
                          .type(String.class)
                          .desc("SDKMS server endpoint (e.g. " +
                                "https://apps.smartkey.io")
                          .build());

        options.addOption(Option.builder("k").
                          longOpt("keystore")
                          .required(true)
                          .hasArg()
                          .type(String.class)
                          .desc("Path to PKCS#12 keystore with client cert " +
                                "and client key")
                          .build());

        options.addOption(Option.builder("p")
                          .longOpt("keystore-password")
                          .hasArg()
                          .type(String.class)
                          .desc("Password protecting the keystore")
                          .build());

        options.addOption(Option.builder("t")
                          .longOpt("truststore")
                          .hasArg()
                          .type(String.class)
                          .desc("Java truststore containing server certificate")
                          .build());

        options.addOption(Option.builder("d")
                          .longOpt("debug")
                          .desc("Enable client API library debugging")
                          .build());

        options.addOption(Option.builder("D")
                          .longOpt("debug-tls")
                          .desc("Enable Java TLS debugging")
                          .build());

        options.addOption(Option.builder("h")
                          .longOpt("help")
                          .desc("Get usage information and exit")
                          .build());

        CommandLineParser parser = new DefaultParser();
        try {
            CommandLine cmdline = parser.parse(options, args);

            if (cmdline.hasOption("help")) {
                usage(options);
                System.exit(0);
            }

            debug = cmdline.hasOption("debug");
            debug_tls = cmdline.hasOption("debug-tls");
            app_id = cmdline.getOptionValue("app-id");
            if (cmdline.hasOption("server")) {
                server = cmdline.getOptionValue("server");
            } else {
                server = "https://apps.smartkey.io";
            }
            keystore = cmdline.getOptionValue("keystore");
            if (cmdline.hasOption("keystore-password")) {
                keystore_password = cmdline.getOptionValue("keystore-password");
            } else {
                keystore_password = "";
            }
            if (cmdline.hasOption("truststore")) {
                truststore = cmdline.getOptionValue("truststore");
            }
        } catch (ParseException e) {
            usage(options);
            System.out.println("Command-line parsing failed: " +
                               e.getMessage());
            System.exit(1);
        }
    }

    public static void main(String[] args) throws ApiException {
        int retval = 0;
        ClientCertificate client = new ClientCertificate();
        client.parseCmdline(args);
        client.connect();
        client.listSecurityObjects();
    }
}
